/* Compile and link with  gcc Skanna.c -o RunImage -lasound -ludev -lm  */
/* You will need to have installed the development packages for udev, alsa, and usb */
#include <stdio.h>
#include <time.h>
#include <math.h>
#include "hidraw.c"
#include "fcd.c"
#include "alsa/asoundlib.h"

#define ALSA_PCM_NEW_HW_PARAMS_API

int alsa_left,alsa_right,alsa_byte,alsa_icount,alsa_grab,alsa_clipped;
int offset_left,offset_right;

long lar[20001], rar[20001];  /* were 9601 long */

double alsa_rms_left,alsa_rms_right,alsa_rms_vect;
double alsa_rms_left_cor,alsa_rms_right_cor,alsa_rms_vect_cor;
double alsa_dc_left,alsa_dc_right, alsa_norm, alsa_dc_norm;
double dcer_left,dcer_right,ave_left,ave_right;

FILE* outfile;

int alsa_rc;
int alsa_size;
snd_pcm_t *alsa_handle;
snd_pcm_hw_params_t *alsa_params;
unsigned int alsa_val;
int alsa_dir;
snd_pcm_uframes_t alsa_frames;
char *alsa_buffer;
char cback,gainval;

unsigned int fstart,fstop,f,df,ftuned;
double dwell_time,tnow,tstarted;
int i,iloop, mhz,khz,ok,iwait,iwait2;

char appdirname[1024];
char settingsfilename[1024];
char outfilename[1024];
char charzero,BTchar,LNAchar,MGchar;
char AlsaDevString[128];

void user_input(void);
void setup_cube(void);
void setup_sound(void);
void grab_sound(void);
void shutdown_sound(void);
int read_settingsfile(void);


hid_device* dev_fcpp;


int main(int argc,char*argv[])
{
  
  printf(" FUN Sweeper V2.00 Jim Lesurf 1st May 2019\n");
  printf(" ==========================================\n\n");
  
  printf(" Simple ROX RF sweep scanner for FUNCube pro+\n\n");
  
  mhz=1000000;  khz=1000; dwell_time=0.5;
  fstart=400*mhz; fstop=900*mhz; df=200*khz;
  
  if (argc>0)
  {
     sprintf(appdirname,"%s",argv[0]);
     i=(int)strrchr(appdirname,'/')-(int)appdirname;
     appdirname[i]=charzero;
     printf("Application Dir at = %s\n\n",appdirname);
     sprintf(settingsfilename,"%s/settings",appdirname);
     printf("Settings file = %s\n",settingsfilename);    
     sprintf(outfilename,"%s/Output/FUNskan_output.csv",appdirname);
     printf("Output to %s\n",outfilename);  
     ok=read_settingsfile();
  }
  
  setup_cube();
  
  user_input();
  
 
  
  
  
  if(ok==1)
  {
    outfile=fopen(outfilename,"wt");
    f=fstart;
    setup_sound();
    printf("start f = %d\n",f);
    dcer_left=0.0; dcer_right=0.0;
    iloop=0; 
    do
    { 
      fcdAppSetFreq(f,&ftuned);
      if(f==ftuned)
      {
        grab_sound();
      }
      else
      {
        printf("Frequency not tuned %d %d\n",f,ftuned);
      }
  
      f+=df;
      iloop++;
    
  
    } while (f<=fstop);
  
    shutdown_sound();
  
    fclose(outfile);
  
    dcer_left=dcer_left/(double)iloop;
    dcer_right=dcer_right/(double)iloop;
  
    printf("Average dc offsets %8.4lf %8.4lf\n",dcer_left,dcer_right);
  
    if(fcdGetMode()!=FCD_MODE_APP)
    {
      printf("Not App mode at finish!\n");
      
    }
   
  }
  
  sprintf(outfilename,"%s/dump_grab.csv",appdirname);
  
  outfile=fopen(outfilename,"wt");
  iloop=0;
  do
  {
    fprintf(outfile,"%5d, %5ld, %5ld,\n",iloop,lar[iloop],rar[iloop]);
    iloop++;
  } while (iloop<9600);
  
  fclose(outfile);
  
  printf("   *  Done!  *\n");
}

void setup_cube(void)
{
  
  if(fcdGetMode()!=FCD_MODE_APP)
  {
    printf("Not App mode!\n");
    ok=0;
  }
  
  if(ok==1)
  {
    
    if(BTchar=='Y')
    {
      fcdAppSetBiasTee(1);
    }
    else
    {
      fcdAppSetBiasTee(0);
    }
    
    fcdAppGetBiasTee(&cback);    
    if(cback==1)
    {
      printf(" * * Bias Tee ON! * *\n");
    }
    else
    {
      printf("Bias Tee OFF\n");
    }
    
      
    if(LNAchar=='Y')
    {
      fcdAppSetLna(1);
    }
    else
    {
      fcdAppSetLna(0);
    }
     
    fcdAppGetLna(&cback);
    if(cback==1)
    {
      printf("Lna ON\n");
    }
    else
    {
      printf("Lna OFF");
    }
        
    if(MGchar=='Y')
    {
      fcdAppSetMixerGain(1);
    }
    else
    {
      fcdAppSetMixerGain(0);
    }
    
        
    fcdAppGetMixerGain(&cback);
    if(cback==1)
    {
      printf("Mixer Gain ON\n");
    }
    else
    {
      printf("Mixer Gain OFF\n");
    }
    
    
    
    
  }
 
}

int read_settingsfile(void)
{
  char tempname[128];
  FILE* settingsfile;
   int rok,lok;
   settingsfile=fopen(settingsfilename,"rt");

   if((int)settingsfile>0)
   {
     printf("\nReading settings ok\n");
     rok=1;
     fscanf(settingsfile,"%d %d\n",&offset_left,&offset_right);
     printf("Offsets  %d  %d\n",offset_left,offset_right);
     fscanf(settingsfile,"%c %c %c",&LNAchar,&MGchar,&BTchar);
     printf("Controls: LNA = %c   MixerGain = %c  BiasTee = %c\n",LNAchar,MGchar,BTchar);
     fscanf(settingsfile,"%s",AlsaDevString);
     printf("FUncube ALSA device = %s\n",AlsaDevString);
     fscanf(settingsfile,"%d",&df);
     df=df*khz;
     printf("Frequency step = %d Hz\n",df);
     
   }
   else
   {
     printf("no settings file!\n");
     rok=0;
   }
   fclose(settingsfile);
   return rok;
  
}

void user_input(void)
{

  double uin;
  int uii;
 
  
  printf("Enter frequency to start sweep [MHz] => ");
  scanf("%lf",&uin);
  fstart=(unsigned int)(mhz*uin);
  
  
  printf("Enter frequency to end sweep [MHz] => ");
  scanf("%lf",&uin);
  
  if((mhz*uin)>fstart)
  {
    fstop=(unsigned int)(mhz*uin);
  }
  else
  {
    fstop=fstart;
    fstart=(unsigned int)(mhz*uin);
  }
  
  printf("Will scan %d to %d Hz\n",fstart,fstop);
  
  printf("\nEnter IF/Mixer Gain [dB] => ");
  scanf("%d",&uii);
  
  if(uii<0) uii=0;
  if(uii>59) uii=59;
  gainval=(char)uii;
  
  fcdAppSetIfGain(gainval);
  fcdAppGetIfGain(&cback);
  if(cback!=gainval)
  {
    printf("IF gain error\n");
    ok=0;
  }
  else
  {
    printf("Mixer gain = %d dB\n",(int)gainval);
  }
  
}

void grab_sound(void)
{
  int ipair;
  short sleft,sright;
  short* spointer;
  alsa_rms_left=0.0;  alsa_rms_right=0.0; alsa_rms_vect=0.0;
  alsa_rms_left_cor=0.0;  alsa_rms_right_cor=0.0; alsa_rms_vect_cor=0.0;
  alsa_dc_left=0.0;  alsa_dc_right=0.0;
  ave_left=0.0; ave_right=0.0;
  alsa_rc = snd_pcm_readi(alsa_handle, alsa_buffer, alsa_frames);
    if (alsa_rc == -EPIPE) {
      /* EPIPE means overrun */
      fprintf(stderr, "overrun occurred\n");

      snd_pcm_prepare(alsa_handle);
    } else if (alsa_rc < 0) {
      fprintf(stderr,
              "error from read: %s\n", 
              snd_strerror(alsa_rc));
    } else if (alsa_rc != (int)alsa_frames) {
      fprintf(stderr, "short read, read %d frames\n", alsa_rc);
    }
    
  
  spointer=(short*)alsa_buffer;
  
  alsa_icount=0; ipair=0;
  do
  {
    /*
    alsa_left=(int)((unsigned int)alsa_buffer[alsa_icount]+256*(unsigned int)alsa_buffer[alsa_icount+1]);
    alsa_right=(int)((unsigned int)alsa_buffer[alsa_icount+2]+256*(unsigned int)alsa_buffer[alsa_icount+3]);
    */
    
    sleft=spointer[alsa_icount]; sright=spointer[alsa_icount+1];
    
    alsa_left=(int)sleft; alsa_right=(int)sright;
    
    if(alsa_left>=32768) 
    {
      alsa_left-=65536;
    }
    if(alsa_right>=32768)
    {
      alsa_right-=65536;
    }
    
    alsa_left-=offset_left;
    alsa_right-=offset_right;
    
    ave_left+=alsa_left;
    ave_right+=alsa_right;
    
    lar[ipair]=alsa_left;  rar[ipair]=alsa_right;
    
    alsa_icount+=2;
    ipair++;
    
  } while (ipair<19200);
  
  ave_left=ave_left/19200.0; ave_right=ave_right/19200.0;
  
  ipair=0;
  do
  {
    alsa_left=lar[ipair];  alsa_right=rar[ipair];
  
    alsa_dc_left+=alsa_left; alsa_dc_right+=alsa_right;
    alsa_rms_left+=(double)(alsa_left*alsa_left);  alsa_rms_right+=(double)(alsa_right*alsa_right);
    alsa_rms_vect+=(double)( alsa_left*alsa_left + alsa_right*alsa_right);
    
    alsa_left-=ave_left; alsa_right-=ave_right;
    
    alsa_rms_left_cor+=(double)(alsa_left*alsa_left);  alsa_rms_right_cor+=(double)(alsa_right*alsa_right);
    alsa_rms_vect_cor+=(double)( alsa_left*alsa_left + alsa_right*alsa_right);
    
    ipair++;
  } while (ipair<19200);
  
  
  
  alsa_rms_right=10.0*log10(alsa_rms_right*alsa_norm);
  alsa_rms_left=10.0*log10(alsa_rms_left*alsa_norm);
  alsa_rms_vect=10.0*log10(alsa_rms_vect*alsa_norm);
  
  alsa_rms_right_cor=10.0*log10(alsa_rms_right_cor*alsa_norm);
  alsa_rms_left_cor=10.0*log10(alsa_rms_left_cor*alsa_norm);
  alsa_rms_vect_cor=10.0*log10(alsa_rms_vect_cor*alsa_norm);
  
  alsa_dc_left=alsa_dc_left*alsa_dc_norm;
  
  alsa_dc_right=alsa_dc_right*alsa_dc_norm;
  
  
  printf("f= %d  %6.2lf  %6.2lf %6.2lf  %6.2lf %6.2lf  %6.2lf\n", f, alsa_rms_left, alsa_rms_right, alsa_rms_left_cor, alsa_rms_right_cor, alsa_dc_left, alsa_dc_right);
  
  dcer_left+=alsa_dc_left; dcer_right+=alsa_dc_right;
  
  
  fprintf(outfile,"%9.4f,  %8.4lf,  %8.4lf\n",0.000001*(double)f,alsa_rms_vect,alsa_rms_vect_cor);
}

void shutdown_sound(void)
{
  snd_pcm_drain(alsa_handle);
  snd_pcm_close(alsa_handle);
  free(alsa_buffer);
}

void setup_sound(void)
{
  
  /* Open PCM device for recording (capture). */
  alsa_rc = snd_pcm_open(&alsa_handle, AlsaDevString,
                    SND_PCM_STREAM_CAPTURE, 0);
  if (alsa_rc < 0) {
    fprintf(stderr, 
            "unable to open pcm device: %s\n",
            snd_strerror(alsa_rc));
    exit(1);
  }
  
  /* Allocate a hardware parameters object. */
  snd_pcm_hw_params_alloca(&alsa_params);

  /* Fill it in with default values. */
  snd_pcm_hw_params_any(alsa_handle, alsa_params);

  /* Set the desired hardware parameters. */

  /* Interleaved mode */
  snd_pcm_hw_params_set_access(alsa_handle, alsa_params,
                      SND_PCM_ACCESS_RW_INTERLEAVED);

  /* Signed 16-bit little-endian format */
  snd_pcm_hw_params_set_format(alsa_handle, alsa_params,
                              SND_PCM_FORMAT_S16_LE);

  /* Two channels (stereo) */
  snd_pcm_hw_params_set_channels(alsa_handle, alsa_params, 2);

  /* bits/second sampling rate */
  alsa_val = 192000;
  snd_pcm_hw_params_set_rate_near(alsa_handle, alsa_params, 
                                  &alsa_val, &alsa_dir);

  /* Set period size to 128 frames. i.e. 19200 value */
  alsa_frames = 19200;  
  snd_pcm_hw_params_set_period_size_near(alsa_handle, 
                              alsa_params, &alsa_frames, &alsa_dir);
  
  alsa_norm=1.0/(double)(alsa_frames*32768);
  alsa_norm=alsa_norm/32768.0;
  alsa_dc_norm=1.0/(double)alsa_frames;
  
  /* Write the parameters to the driver */
  alsa_rc = snd_pcm_hw_params(alsa_handle, alsa_params);
  if (alsa_rc < 0) {
    fprintf(stderr,
            "unable to set hw parameters: %s\n",
            snd_strerror(alsa_rc));
    exit(1);
  }

  /* Use a buffer large enough to hold one period */
  snd_pcm_hw_params_get_period_size(alsa_params,
                                      &alsa_frames, &alsa_dir);
  alsa_size = alsa_frames * 4; /* 2 bytes/sample, 2 channels */
  alsa_buffer = (char *) malloc(alsa_size);

  
  snd_pcm_hw_params_get_period_time(alsa_params,
                                         &alsa_val, &alsa_dir);
  
  
}


